@@ -36,10 +36,9 @@
 #include <bluetooth-chooser.h>
 #include <bluetooth-agent.h>
 
-#include "helper.h"
+#include "pin.h"
 
 #define AGENT_PATH "/org/bluez/agent/wizard"
-#define PIN_CODE_DB "pin-code-database.txt"
 
 static gboolean set_page_search_complete(void);
 
@@ -54,10 +53,11 @@ static gboolean target_ssp = FALSE;
 
 /* NULL means automatic, anything else is a pincode specified by the user */
 static gchar *user_pincode = NULL;
+/* If TRUE, then we won't display the PIN code to the user when pairing */
+static gboolean automatic_pincode = FALSE;
 
 static GtkWidget *window_assistant = NULL;
 static GtkWidget *page_search = NULL;
-static GtkWidget *passkey_button = NULL;
 static GtkWidget *page_setup = NULL;
 static GtkWidget *page_summary = NULL;
 
@@ -67,93 +67,13 @@ static GtkWidget *label_passkey_help = NULL;
 
 static BluetoothChooser *selector = NULL;
 
-#define TYPE_IS(x, r) {				\
-	if (g_str_equal(type, x)) return r;	\
-}
-
-static guint string_to_type(const char *type)
-{
-	TYPE_IS ("any", BLUETOOTH_TYPE_ANY);
-	TYPE_IS ("mouse", BLUETOOTH_TYPE_MOUSE);
-	TYPE_IS ("keyboard", BLUETOOTH_TYPE_KEYBOARD);
-	TYPE_IS ("headset", BLUETOOTH_TYPE_HEADSET);
-	TYPE_IS ("headphones", BLUETOOTH_TYPE_HEADPHONES);
-	TYPE_IS ("audio", BLUETOOTH_TYPE_OTHER_AUDIO);
-	TYPE_IS ("printer", BLUETOOTH_TYPE_PRINTER);
-	TYPE_IS ("network", BLUETOOTH_TYPE_NETWORK);
-
-	g_warning ("unhandled type '%s'", type);
-	return BLUETOOTH_TYPE_ANY;
-}
-
-static char *set_pincode_for_device(guint type, const char *address, const char *name)
-{
-	char *contents, **lines;
-	char *ret_pin = NULL;
-	guint i;
-
-	/* Load the PIN database and split it in lines */
-	if (!g_file_get_contents(PIN_CODE_DB, &contents, NULL, NULL)) {
-		char *filename;
-
-		filename = g_build_filename(PKGDATADIR, PIN_CODE_DB, NULL);
-		if (!g_file_get_contents(filename, &contents, NULL, NULL)) {
-			g_warning("Could not load "PIN_CODE_DB);
-			g_free (filename);
-			return NULL;
-		}
-		g_free (filename);
-	}
-
-	lines = g_strsplit(contents, "\n", -1);
-	g_free (contents);
-	if (lines == NULL) {
-		g_warning("Could not parse "PIN_CODE_DB);
-		return NULL;
-	}
-
-	/* And now process each line */
-	for (i = 0; lines[i] != NULL; i++) {
-		char **items;
-		guint ltype;
-		const char *laddress, *lname, *lpin;
-
-		/* Ignore comments and empty lines */
-		if (lines[i][0] == '#' || lines[i][0] == '\0')
-			continue;
-
-		items = g_strsplit(lines[i], "\t", 4);
-		ltype = string_to_type(items[0]);
-
-		if (ltype != BLUETOOTH_TYPE_ANY && ltype != type) {
-			g_strfreev (items);
-			continue;
-		}
-		laddress = items[1];
-		lname = items[2];
-		lpin = items[3];
-
-		/* If we have an address, does the OUI prefix match? */
-		if (strlen(laddress) > 0 && g_str_has_prefix(address, laddress) == FALSE) {
-			g_strfreev (items);
-			continue;
-		}
-
-		/* If we have a name, does it match? */
-		if (strlen(lname) > 0 && g_str_equal(name, lname) == FALSE) {
-			g_strfreev (items);
-			continue;
-		}
-
-		/* Everything matches, we have a pincode */
-		ret_pin = g_strdup(lpin);
-		g_strfreev(items);
-		break;
-	}
-
-	g_strfreev(lines);
-	return ret_pin;
-}
+static GtkWidget *passkey_dialog = NULL;
+static GtkWidget *radio_auto = NULL;
+static GtkWidget *radio_0000 = NULL;
+static GtkWidget *radio_1111 = NULL;
+static GtkWidget *radio_1234 = NULL;
+static GtkWidget *radio_custom = NULL;
+static GtkWidget *entry_custom = NULL;
 
 static void
 set_large_label (GtkLabel *label, const char *text)
@@ -170,16 +90,36 @@ static gboolean pincode_callback(DBusGMethodInvocation *context,
 {
 	char *pincode;
 
-	if (user_pincode != NULL && strlen(user_pincode) == 4) {
+	if (user_pincode != NULL && *user_pincode != '\0') {
 		pincode = g_strdup (user_pincode);
 	} else {
-		pincode = set_pincode_for_device(target_type, target_address, target_name);
+		pincode = get_pincode_for_device(target_type, target_address, target_name);
 		if (pincode == NULL)
 			pincode = g_strdup(target_pincode);
+		else
+			automatic_pincode = TRUE;
 	}
 
-	gtk_label_set_markup(GTK_LABEL(label_passkey_help), _("Please enter the following passkey:"));
-	set_large_label (GTK_LABEL (label_passkey), pincode);
+	if (automatic_pincode == FALSE) {
+		gtk_widget_show (label_passkey_help);
+		gtk_widget_show (label_passkey);
+
+		gtk_label_set_markup(GTK_LABEL(label_passkey_help), _("Please enter the following passkey:"));
+		set_large_label (GTK_LABEL (label_passkey), pincode);
+	} else {
+		char *text;
+
+		gtk_widget_show (label_passkey_help);
+		gtk_widget_hide (label_passkey);
+
+		/* translators:
+		 * The '%s' is the device name, for example:
+		 * Please wait whilst 'Sony Bluetooth Headset' is being paired
+		 */
+		text = g_strdup_printf(_("Please wait whilst '%s' is being paired"), target_name);
+		gtk_label_set_markup(GTK_LABEL(label_passkey_help), text);
+		g_free(text);
+	}
 
 	dbus_g_method_return(context, pincode);
 	g_free(pincode);
@@ -207,6 +147,8 @@ static gboolean display_callback(DBusGMethodInvocation *context,
 		done = g_strdup ("");
 	}
 
+	gtk_widget_show (label_passkey_help);
+
 	gtk_label_set_markup(GTK_LABEL(label_passkey_help), _("Please enter the following passkey:"));
 	text = g_strdup_printf("%s%s", done, code + entered);
 	set_large_label (GTK_LABEL (label_passkey), text);
@@ -227,17 +169,27 @@ static gboolean cancel_callback(DBusGMethodInvocation *context,
 {
 	gchar *text;
 
-	if (target_ssp == FALSE)
-		text = g_strdup_printf(_("Pairing with %s canceled"),
-								target_name);
-	else
-		text = g_strdup_printf(_("Pairing with %s finished"),
-								target_name);
+	if (target_ssp == FALSE) {
+		/* translators:
+		 * The '%s' is the device name, for example:
+		 * Pairing with 'Sony Bluetooth Headset' cancelled
+		 */
+		text = g_strdup_printf(_("Pairing with '%s' cancelled"), target_name);
+	} else {
+		/* translators:
+		 * The '%s' is the device name, for example:
+		 * Pairing with 'Sony Bluetooth Headset' finished
+		 */
+		text = g_strdup_printf(_("Pairing with '%s' finished"), target_name);
+	}
 
 	gtk_label_set_markup(GTK_LABEL(label_setup), text);
 
 	g_free(text);
 
+	gtk_widget_hide (label_passkey);
+	gtk_widget_hide (label_passkey_help);
+
 	gtk_label_set_markup(GTK_LABEL(label_passkey_help), NULL);
 	gtk_label_set_markup(GTK_LABEL(label_passkey), NULL);
 
@@ -248,6 +200,10 @@ static gboolean cancel_callback(DBusGMethodInvocation *context,
 
 static void connect_callback(gpointer user_data)
 {
+	GtkAssistant *assistant = user_data;
+
+	gtk_widget_hide (label_passkey_help);
+	gtk_assistant_set_page_complete(assistant, page_setup, TRUE);
 }
 
 static void create_callback(const char *path, gpointer user_data)
@@ -259,29 +215,37 @@ static void create_callback(const char *path, gpointer user_data)
 	if (path != NULL) {
 		gint page;
 
-		text = g_strdup_printf(_("Successfully paired with %s"),
-								target_name);
+		/* translators:
+		 * The '%s' is the device name, for example:
+		 * Successfully paired with 'Sony Bluetooth Headset'
+		 */
+		text = g_strdup_printf(_("Successfully paired with '%s'"), target_name);
 
 		page = gtk_assistant_get_current_page(assistant);
 		if (page > 0)
 			gtk_assistant_set_current_page(assistant, page + 1);
 
-		if (target_type == BLUETOOTH_TYPE_KEYBOARD ||
-					target_type == BLUETOOTH_TYPE_MOUSE) {
-			bluetooth_client_set_trusted(client, path, TRUE);
-			bluetooth_client_connect_input(client, path,
-						connect_callback, assistant);
-		}
+		bluetooth_client_set_trusted(client, path, TRUE);
 
-		complete = TRUE;
+		if (bluetooth_client_connect_service(client, path, connect_callback, assistant) != FALSE) {
+			gtk_label_set_text (GTK_LABEL (label_passkey_help),
+					    _("Please wait while setting up the device..."));
+			gtk_widget_show (label_passkey_help);
+		} else {
+			complete = TRUE;
+		}
 	} else {
-		text = g_strdup_printf(_("Pairing with %s failed"),
-								target_name);
+		/* translators:
+		 * The '%s' is the device name, for example:
+		 * Pairing with 'Sony Bluetooth Headset' failed
+		 */
+		text = g_strdup_printf(_("Pairing with '%s' failed"), target_name);
 	}
 
 	gtk_label_set_markup(GTK_LABEL(label_setup), text);
-	gtk_label_set_markup(GTK_LABEL(label_passkey_help), NULL);
-	gtk_label_set_markup(GTK_LABEL(label_passkey), NULL);
+
+	gtk_widget_hide (label_passkey_help);
+	gtk_widget_hide (label_passkey);
 
 	g_free(text);
 
@@ -296,25 +260,20 @@ static void close_callback(GtkWidget *assistant, gpointer data)
 }
 
 static void prepare_callback(GtkWidget *assistant,
-					GtkWidget *page, gpointer data)
+                             GtkWidget *page, gpointer data)
 {
 	gboolean complete = TRUE;
 	const char *path = AGENT_PATH;
 
 	if (page == page_search) {
-		gtk_assistant_add_action_widget (GTK_ASSISTANT (assistant), passkey_button);
 		complete = set_page_search_complete ();
 		bluetooth_client_start_discovery(client);
 	} else {
-		if (gtk_widget_get_parent (passkey_button) != NULL) {
-			g_object_ref (passkey_button);
-			gtk_assistant_remove_action_widget (GTK_ASSISTANT (assistant), passkey_button);
-		}
 		bluetooth_client_stop_discovery(client);
 	}
 
 	if (page == page_setup) {
-		gchar *text, *markup, *address, *name;
+		gchar *text, *address, *name, *pincode;
 		guint type;
 
 		/* Get the info about the device now,
@@ -333,29 +292,24 @@ static void prepare_callback(GtkWidget *assistant,
 
 		target_type = type;
 
-		text = g_strdup_printf(_("Connecting to %s now ..."),
-								target_name);
-		markup = g_strdup_printf("%s\n\n", text);
+		/* translators:
+		 * The '%s' is the device name, for example:
+		 * Connecting to 'Sony Bluetooth Headset' now...
+		 */
+		text = g_strdup_printf(_("Connecting to '%s' now..."), target_name);
+		gtk_label_set_markup(GTK_LABEL(label_setup), text);
+
 		g_free(text);
 
-		gtk_label_set_markup(GTK_LABEL(label_setup), markup);
-
-		g_free(markup);
-
 		complete = FALSE;
 
 		g_object_ref(agent);
 
-		if (target_type == BLUETOOTH_TYPE_MOUSE)
-			path = NULL;
-
-		/* Sony PlayStation 3 Remote Control */
-		if (g_str_equal(target_name, "BD Remote Control") == TRUE &&
-					(g_str_has_prefix(target_address,
-							"00:19:C1:") == TRUE ||
-					g_str_has_prefix(target_address,
-							"00:1E:3D:") == TRUE))
+		/* Do we pair, or don't we? */
+		pincode = get_pincode_for_device (target_type, target_address, target_name);
+		if (pincode != NULL && g_str_equal (pincode, "NULL"))
 			path = NULL;
+		g_free (pincode);
 
 		bluetooth_client_create_device(client, target_address,
 					path, create_callback, assistant);
@@ -365,79 +319,6 @@ static void prepare_callback(GtkWidget *assistant,
 							page, complete);
 }
 
-static GtkWidget *create_vbox(GtkWidget *assistant, GtkAssistantPageType type,
-				const gchar *title, const gchar *section)
-{
-	GtkWidget *vbox;
-	GtkWidget *label;
-	GdkPixbuf *pixbuf;
-	gchar *str;
-
-	vbox = gtk_vbox_new(FALSE, 6);
-
-	gtk_container_set_border_width(GTK_CONTAINER(vbox), 24);
-	gtk_assistant_append_page(GTK_ASSISTANT(assistant), vbox);
-	gtk_assistant_set_page_type(GTK_ASSISTANT(assistant), vbox, type);
-	gtk_assistant_set_page_title(GTK_ASSISTANT(assistant), vbox, title);
-
-	pixbuf = gtk_icon_theme_load_icon(gtk_icon_theme_get_default(),
-						"bluetooth", 32, 0, NULL);
-	gtk_assistant_set_page_header_image(GTK_ASSISTANT(assistant),
-								vbox, pixbuf);
-	g_object_unref(pixbuf);
-
-	if (section) {
-		label = gtk_label_new(NULL);
-		str = g_strdup_printf("<b>%s</b>\n", section);
-		gtk_label_set_markup(GTK_LABEL(label), str);
-		g_free(str);
-		gtk_misc_set_alignment(GTK_MISC(label), 0, 0);
-		gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
-	}
-
-	return vbox;
-}
-
-static void create_intro(GtkWidget *assistant)
-{
-	GtkWidget *vbox;
-	GtkWidget *label;
-	GtkWidget *combo;
-	GtkTreeModel *model;
-	GtkCellRenderer *renderer;
-
-	vbox = create_vbox(assistant, GTK_ASSISTANT_PAGE_INTRO,
-			_("Introduction"),
-			_("Welcome to the Bluetooth device setup wizard"));
-
-	label = gtk_label_new(NULL);
-	gtk_label_set_markup(GTK_LABEL(label), _("The device wizard will "
-				"walk you through the process of configuring "
-				"Bluetooth enabled devices for use with this "
-				"computer.\n\n"));
-	gtk_misc_set_alignment(GTK_MISC(label), 0, 0);
-	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
-	gtk_widget_set_size_request(GTK_WIDGET(label), 380, -1);
-	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, TRUE, 0);
-
-	combo = gtk_combo_box_new();
-
-	model = bluetooth_client_get_adapter_model(client);
-	gtk_combo_box_set_model(GTK_COMBO_BOX(combo), model);
-	g_object_unref(model);
-
-	gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0);
-	gtk_cell_layout_clear(GTK_CELL_LAYOUT(combo));
-
-	renderer = gtk_cell_renderer_text_new();
-	gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), renderer, TRUE);
-	gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo), renderer,
-					"text", BLUETOOTH_COLUMN_NAME, NULL);
-
-	if (gtk_tree_model_iter_n_children(model, NULL) > 1)
-		gtk_box_pack_start(GTK_BOX(vbox), combo, FALSE, FALSE, 0);
-}
-
 static gboolean
 set_page_search_complete (void)
 {
@@ -461,8 +342,7 @@ set_page_search_complete (void)
 	return complete;
 }
 
-static gboolean entry_custom_event(GtkWidget *entry,
-					GdkEventKey *event, GtkWidget *box)
+static gboolean entry_custom_event(GtkWidget *entry, GdkEventKey *event)
 {
 	if (event->length == 0)
 		return FALSE;
@@ -474,10 +354,10 @@ static gboolean entry_custom_event(GtkWidget *entry,
 	return TRUE;
 }
 
-static void entry_custom_changed(GtkWidget *entry, GtkWidget *dialog)
+static void entry_custom_changed(GtkWidget *entry)
 {
 	user_pincode = g_strdup (gtk_entry_get_text(GTK_ENTRY(entry)));
-	gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog),
+	gtk_dialog_set_response_sensitive (GTK_DIALOG (passkey_dialog),
 					   GTK_RESPONSE_ACCEPT,
 					   gtk_entry_get_text_length (GTK_ENTRY (entry)) >= 1);
 }
@@ -488,9 +368,13 @@ static void toggle_set_sensitive(GtkWidget *button, GtkWidget *widget)
 
 	active = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button));
 	gtk_widget_set_sensitive(widget, active);
+	/* When selecting another PIN, make sure the "Close" button is sensitive */
+	if (!active)
+		gtk_dialog_set_response_sensitive (GTK_DIALOG (passkey_dialog),
+						   GTK_RESPONSE_ACCEPT, TRUE);
 }
 
-static void set_user_pincode(GtkWidget *button, GtkWidget *dialog)
+static void set_user_pincode(GtkWidget *button)
 {
 	GSList *list, *l;
 
@@ -506,7 +390,7 @@ static void set_user_pincode(GtkWidget *button, GtkWidget *dialog)
 		/* Is it radio_fixed that changed? */
 		radio = g_object_get_data (G_OBJECT (button), "button");
 		if (radio != NULL) {
-			set_user_pincode (radio, NULL);
+			set_user_pincode (radio);
 			return;
 		}
 
@@ -516,7 +400,7 @@ static void set_user_pincode(GtkWidget *button, GtkWidget *dialog)
 		if (entry != NULL) {
 			g_free (user_pincode);
 			user_pincode = g_strdup (gtk_entry_get_text(entry));
-			gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog),
+			gtk_dialog_set_response_sensitive (GTK_DIALOG (passkey_dialog),
 							   GTK_RESPONSE_ACCEPT,
 							   gtk_entry_get_text_length (entry) >= 1);
 		} else if (pin != NULL) {
@@ -548,80 +432,15 @@ static void select_device_changed(BluetoothChooser *selector,
 static void
 passkey_option_button_clicked (GtkButton *button, gpointer data)
 {
-	GtkWidget *dialog;
-	GtkWidget *vbox;
-	GtkWidget *radio_auto;
-	GtkWidget *radio_fixed;
-	GtkWidget *align_fixed;
-	GtkWidget *vbox_fixed;
 	GtkWidget *radio;
-	GtkWidget *radio_0000;
-	GtkWidget *radio_1111;
-	GtkWidget *radio_1234;
-	GtkWidget *hbox_custom;
-	GtkWidget *radio_custom;
-	GtkWidget *entry_custom;
 	char *oldpin;
 
 	oldpin = user_pincode;
 	user_pincode = NULL;
 
-	dialog = gtk_dialog_new_with_buttons (_("Passkey options"),
-					      GTK_WINDOW (data),
-					      GTK_DIALOG_MODAL | GTK_DIALOG_NO_SEPARATOR,
-					      GTK_STOCK_CANCEL,
-					      GTK_RESPONSE_REJECT,
-					      GTK_STOCK_CLOSE,
-					      GTK_RESPONSE_ACCEPT,
-					      NULL);
-	vbox = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
-
-	radio_auto = gtk_radio_button_new_with_mnemonic(NULL,
-					_("_Automatic passkey selection"));
-	gtk_container_add(GTK_CONTAINER(vbox), radio_auto);
-
-	radio_fixed = gtk_radio_button_new_with_mnemonic_from_widget(GTK_RADIO_BUTTON(radio_auto), _("Use _fixed passkey:"));
-	gtk_container_add(GTK_CONTAINER(vbox), radio_fixed);
-
-	align_fixed = gtk_alignment_new(0, 0, 1, 1);
-	gtk_alignment_set_padding(GTK_ALIGNMENT(align_fixed), 0, 0, 24, 0);
-	gtk_container_add(GTK_CONTAINER(vbox), align_fixed);
-	vbox_fixed = gtk_vbox_new(FALSE, 0);
-	gtk_container_add(GTK_CONTAINER(align_fixed), vbox_fixed);
-
-	radio_0000 = gtk_radio_button_new_with_label(NULL, _("'0000' (most headsets, mice and GPS devices)"));
-	gtk_container_add(GTK_CONTAINER(vbox_fixed), radio_0000);
-
-	radio_1111 = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(radio_0000), _("'1111'"));
-	gtk_container_add(GTK_CONTAINER(vbox_fixed), radio_1111);
-
-	radio_1234 = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(radio_0000), _("'1234'"));
-	gtk_container_add(GTK_CONTAINER(vbox_fixed), radio_1234);
-
-	hbox_custom = gtk_hbox_new(FALSE, 6);
-	radio_custom = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(radio_0000), _("Custom passkey code:"));
-	entry_custom = gtk_entry_new();
-	gtk_entry_set_max_length (GTK_ENTRY (entry_custom), 20);
-	gtk_entry_set_width_chars(GTK_ENTRY(entry_custom), 20);
-	g_signal_connect (entry_custom, "key-press-event",
-			  G_CALLBACK (entry_custom_event), NULL);
-	g_signal_connect (entry_custom, "changed",
-			  G_CALLBACK (entry_custom_changed), dialog);
-	gtk_box_pack_start(GTK_BOX(hbox_custom), radio_custom,
-			FALSE, FALSE, 0);
-	gtk_box_pack_start(GTK_BOX(hbox_custom), entry_custom,
-			FALSE, FALSE, 0);
-	gtk_container_add(GTK_CONTAINER(vbox_fixed), hbox_custom);
-
-	toggle_set_sensitive(radio_fixed, vbox_fixed);
-	g_signal_connect(radio_fixed, "toggled",
-			G_CALLBACK(toggle_set_sensitive), vbox_fixed);
-	toggle_set_sensitive(radio_custom, entry_custom);
-	g_signal_connect(radio_custom, "toggled",
-			G_CALLBACK(toggle_set_sensitive), entry_custom);
+	gtk_window_present(GTK_WINDOW(passkey_dialog));
 
 	/* When reopening, try to guess where the pincode was set */
-	radio = NULL;
 	if (oldpin == NULL)
 		radio = radio_auto;
 	else if (g_str_equal (oldpin, "0000"))
@@ -633,50 +452,62 @@ passkey_option_button_clicked (GtkButton *button, gpointer data)
 	else
 		radio = radio_custom;
 
-	if (radio != radio_auto)
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio_fixed), TRUE);
 	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio), TRUE);
 	if (radio == radio_custom)
 		gtk_entry_set_text (GTK_ENTRY (entry_custom), oldpin);
 
-	g_object_set_data (G_OBJECT (radio_auto), "pin", "");
-	g_object_set_data (G_OBJECT (radio_fixed), "button", radio_0000);
-	g_object_set_data (G_OBJECT (radio_0000), "pin", "0000");
-	g_object_set_data (G_OBJECT (radio_1111), "pin", "1111");
-	g_object_set_data (G_OBJECT (radio_1234), "pin", "1234");
-	g_object_set_data (G_OBJECT (radio_custom), "entry", entry_custom);
-
-	g_signal_connect(radio_auto, "toggled",
-			 G_CALLBACK(set_user_pincode), dialog);
-	g_signal_connect(radio_fixed, "toggled",
-			 G_CALLBACK(set_user_pincode), dialog);
-	g_signal_connect(radio_0000, "toggled",
-			 G_CALLBACK(set_user_pincode), dialog);
-	g_signal_connect(radio_1111, "toggled",
-			 G_CALLBACK(set_user_pincode), dialog);
-	g_signal_connect(radio_1234, "toggled",
-			 G_CALLBACK(set_user_pincode), dialog);
-	g_signal_connect(radio_custom, "toggled",
-			 G_CALLBACK(set_user_pincode), dialog);
-
-	gtk_widget_show_all (vbox);
-
-	if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_ACCEPT) {
+	if (gtk_dialog_run (GTK_DIALOG (passkey_dialog)) != GTK_RESPONSE_ACCEPT) {
 		g_free (user_pincode);
 		user_pincode = oldpin;
 	} else {
 		g_free (oldpin);
 	}
 
-	gtk_widget_destroy (dialog);
+	gtk_widget_hide(passkey_dialog);
 }
 
-static void create_search(GtkWidget *assistant)
+static GtkWidget *create_wizard(void)
 {
-	GtkWidget *vbox;
-	GtkWidget *button;
+	GtkAssistant *assistant;
+	GtkBuilder *builder;
+	GError *err = NULL;
+	GtkWidget *combo, *page_intro;
+	GtkTreeModel *model;
+	GtkCellRenderer *renderer;
+	GdkPixbuf *pixbuf;
 
-	vbox = create_vbox(assistant, GTK_ASSISTANT_PAGE_CONTENT, _("Device search"), NULL);
+	builder = gtk_builder_new ();
+	if (gtk_builder_add_from_file (builder, "wizard.ui", NULL) == 0) {
+		if (gtk_builder_add_from_file (builder, PKGDATADIR "/wizard.ui", &err) == 0) {
+			g_warning ("Could not load UI from %s: %s", PKGDATADIR "/wizard.ui", err->message);
+			g_error_free(err);
+			return NULL;
+		}
+	}
+
+	assistant = GTK_ASSISTANT(gtk_builder_get_object(builder, "assistant"));
+
+	/* Intro page */
+	combo = gtk_combo_box_new();
+
+	model = bluetooth_client_get_adapter_model(client);
+	gtk_combo_box_set_model(GTK_COMBO_BOX(combo), model);
+	g_object_unref(model);
+
+	gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0);
+	gtk_cell_layout_clear(GTK_CELL_LAYOUT(combo));
+
+	renderer = gtk_cell_renderer_text_new();
+	gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), renderer, TRUE);
+	gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo), renderer,
+					"text", BLUETOOTH_COLUMN_NAME, NULL);
+
+	page_intro = GTK_WIDGET(gtk_builder_get_object(builder, "page_intro"));
+	if (gtk_tree_model_iter_n_children(model, NULL) > 1)
+		gtk_box_pack_start(GTK_BOX(page_intro), combo, FALSE, FALSE, 0);
+
+	/* Search page */
+	page_search = GTK_WIDGET(gtk_builder_get_object(builder, "page_search"));
 
 	/* The selector */
 	selector = BLUETOOTH_CHOOSER (bluetooth_chooser_new(_("Select the device you want to setup")));
@@ -694,72 +525,64 @@ static void create_search(GtkWidget *assistant)
 	g_signal_connect(selector, "notify::device-selected-name",
 			 G_CALLBACK(device_selected_name_cb), assistant);
 
-	gtk_container_add(GTK_CONTAINER(vbox), GTK_WIDGET (selector));
+	gtk_container_add(GTK_CONTAINER(page_search), GTK_WIDGET (selector));
 
-	page_search = vbox;
+	/* Setup page */
+	page_setup = GTK_WIDGET(gtk_builder_get_object(builder, "page_setup"));
+	label_setup = GTK_WIDGET(gtk_builder_get_object(builder, "label_setup"));
+	label_passkey_help = GTK_WIDGET(gtk_builder_get_object(builder, "label_passkey_help"));
+	label_passkey = GTK_WIDGET(gtk_builder_get_object(builder, "label_passkey"));
 
-	button = gtk_button_new_with_mnemonic (_("Passkey _options..."));
-	g_signal_connect (G_OBJECT (button), "clicked",
-			  G_CALLBACK (passkey_option_button_clicked), assistant);
-	gtk_widget_show (button);
+	/* Summary page */
+	page_summary = GTK_WIDGET(gtk_builder_get_object(builder, "page_summary"));
 
-	passkey_button = button;
-}
+	/* Set page icons (named icons not supported by Glade) */
+	pixbuf = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (),
+					   "bluetooth", 32, 0, NULL);
+	gtk_assistant_set_page_header_image (assistant, page_intro, pixbuf);
+	gtk_assistant_set_page_header_image (assistant, page_search, pixbuf);
+	gtk_assistant_set_page_header_image (assistant, page_setup, pixbuf);
+	gtk_assistant_set_page_header_image (assistant, page_summary, pixbuf);
+	g_object_unref (pixbuf);
 
-static void create_setup(GtkWidget *assistant)
-{
-	GtkWidget *vbox;
-	GtkWidget *label;
+	/* Passkey dialog */
+	passkey_dialog = GTK_WIDGET(gtk_builder_get_object(builder, "passkey_dialog"));
+	radio_auto = GTK_WIDGET(gtk_builder_get_object(builder, "radio_auto"));
+	radio_0000 = GTK_WIDGET(gtk_builder_get_object(builder, "radio_0000"));
+	radio_1111 = GTK_WIDGET(gtk_builder_get_object(builder, "radio_1111"));
+	radio_1234 = GTK_WIDGET(gtk_builder_get_object(builder, "radio_1234"));
+	radio_custom = GTK_WIDGET(gtk_builder_get_object(builder, "radio_custom"));
+	entry_custom = GTK_WIDGET(gtk_builder_get_object(builder, "entry_custom"));
 
-	vbox = create_vbox(assistant, GTK_ASSISTANT_PAGE_CONTENT,
-				_("Device setup"),
-				_("Setting up new device"));
+	g_object_set_data (G_OBJECT (radio_auto), "pin", "");
+	g_object_set_data (G_OBJECT (radio_0000), "pin", "0000");
+	g_object_set_data (G_OBJECT (radio_1111), "pin", "1111");
+	g_object_set_data (G_OBJECT (radio_1234), "pin", "1234");
+	g_object_set_data (G_OBJECT (radio_custom), "entry", entry_custom);
 
-	label = gtk_label_new(NULL);
-	gtk_misc_set_alignment(GTK_MISC(label), 0, 0);
-	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
-	gtk_widget_set_size_request(GTK_WIDGET(label), 380, -1);
-	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, TRUE, 0);
-	label_setup = label;
+	g_signal_connect(radio_auto, "toggled",
+			 G_CALLBACK(set_user_pincode), NULL);
+	g_signal_connect(radio_0000, "toggled",
+			 G_CALLBACK(set_user_pincode), NULL);
+	g_signal_connect(radio_1111, "toggled",
+			 G_CALLBACK(set_user_pincode), NULL);
+	g_signal_connect(radio_1234, "toggled",
+			 G_CALLBACK(set_user_pincode), NULL);
+	g_signal_connect(radio_custom, "toggled",
+			 G_CALLBACK(set_user_pincode), NULL);
 
-	label = gtk_label_new(NULL);
-	gtk_misc_set_alignment(GTK_MISC(label), 0, 0);
-	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, TRUE, 0);
-	label_passkey_help = label;
+	g_signal_connect (entry_custom, "key-press-event",
+			  G_CALLBACK (entry_custom_event), NULL);
+	g_signal_connect (entry_custom, "changed",
+			  G_CALLBACK (entry_custom_changed), NULL);
+	g_signal_connect(radio_custom, "toggled",
+			G_CALLBACK(toggle_set_sensitive), entry_custom);
 
-	label = gtk_label_new(NULL);
-	gtk_misc_set_alignment(GTK_MISC(label), 0.5, 0.5);
-	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, TRUE, 0);
-	label_passkey = label;
-
-	page_setup = vbox;
-}
-
-static void create_summary(GtkWidget *assistant)
-{
-	GtkWidget *vbox;
-
-	vbox = create_vbox(assistant, GTK_ASSISTANT_PAGE_SUMMARY,
-				_("Summary"),
-				_("Successfully configured new device"));
-
-	page_summary = vbox;
-}
-
-static GtkWidget *create_wizard(void)
-{
-	GtkWidget *assistant;
-
-	assistant = gtk_assistant_new();
-	gtk_window_set_title(GTK_WINDOW(assistant), _("Bluetooth Device Wizard"));
-	gtk_window_set_position(GTK_WINDOW(assistant), GTK_WIN_POS_CENTER);
-	gtk_window_set_default_size(GTK_WINDOW(assistant), 440, 440);
-
-	create_intro(assistant);
-	//create_type(assistant);
-	create_search(assistant);
-	create_setup(assistant);
-	create_summary(assistant);
+	/* Connect signals.  Autoconnect could be used if the build process
+         * was changed.  Set the GtkBuilder documentation for details.
+         * Check all signals are in the .ui files before removing manual
+         * connections. */
+	/*gtk_builder_connect_signals(builder, NULL);*/
 
 	g_signal_connect(G_OBJECT(assistant), "close",
 					G_CALLBACK(close_callback), NULL);
@@ -767,12 +590,14 @@ static GtkWidget *create_wizard(void)
 					G_CALLBACK(close_callback), NULL);
 	g_signal_connect(G_OBJECT(assistant), "prepare",
 					G_CALLBACK(prepare_callback), NULL);
+	g_signal_connect (gtk_builder_get_object(builder, "passkey_option_button"), "clicked",
+			  G_CALLBACK (passkey_option_button_clicked), assistant);
 
-	gtk_widget_show_all(assistant);
+	gtk_widget_show_all(GTK_WIDGET(assistant));
 
 	gtk_assistant_update_buttons_state(GTK_ASSISTANT(assistant));
 
-	return assistant;
+	return GTK_WIDGET(assistant);
 }
 
 static UniqueResponse
@@ -833,6 +658,8 @@ int main(int argc, char *argv[])
 	bluetooth_agent_setup(agent, AGENT_PATH);
 
 	window = create_wizard();
+	if (window == NULL)
+		return 1;
 	window_assistant = window;
 
 	g_signal_connect (app, "message-received",
@@ -848,3 +675,4 @@ int main(int argc, char *argv[])
 
 	return 0;
 }
+
